/*
 * Copyright (c) 2012-2013, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== SvrDsp.c ========
 *  Platform: OMAPL138_linux_elf
 */

/* this define must precede inclusion of any xdc header file */
#define Registry_CURDESC SvrDsp__Desc
#define MODULE_NAME "SvrDsp"

#include <xdc/std.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Registry.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "SvrDsp.h"
#include "../shared/SvrMsg.h"

/* module structure */
typedef struct {
    MessageQ_Handle localMsgQue;
} SvrDsp_Module;

/* friend data */
Semaphore_Handle SvrDsp_doneFlag = NULL;

/* private data */
Registry_Desc           Registry_CURDESC;
static SvrDsp_Module   Module;


/*
 *  ======== SvrDsp_destroy ========
 */
Int SvrDsp_destroy(Void)
{
    Int status = 0;

    System_printf("--> SvrDsp_destroy:");

    /* TODO enter gate */

    /* delete the local message queue */
    status = MessageQ_delete(&Module.localMsgQue);

    if (status < 0) {
        System_printf("SvrDsp_destroy: message queue delete error=%d \n",
                status);
        goto leave;
    }

    /* delete the done flag */
    Semaphore_delete(&SvrDsp_doneFlag);

leave:
    /* disable log events */
    System_printf("<-- SvrDsp_delete: status=%d \n", status);
    Diags_setMask(MODULE_NAME"-FEX");

    /*
     * Note that there isn't a Registry_removeModule() yet:
     *     https://bugs.eclipse.org/bugs/show_bug.cgi?id=315448
     *
     * ... but this is where we'd call it.
     */

    /* TODO leave gate */
    return(status);
}

/*
 *  ======== SvrDsp_run ========
 */
Int SvrDsp_run(Void)
{
    Int                 status = 0;
    Bool                running = TRUE;
    SvrMsg_Msg *        msg;
    MessageQ_QueueId    queId;
    UInt16              procId;
    UInt16              numProcs;
    UInt16              count = 0;

    System_printf("--> SvrDsp_run:\n");

    numProcs = MultiProc_getNumProcessors();

    while (running) {

        /* wait for inbound message */
        status = MessageQ_get(Module.localMsgQue, (MessageQ_Msg *)&msg,
            MessageQ_FOREVER);

        if (status < 0) {
            System_printf("SvrDsp_run: message get error=%d \n", status);
            goto leave;
        }

        /* process the message */
        queId = MessageQ_getReplyQueue(msg);
        procId = MessageQ_getProcId(queId);
        System_printf("SvrDsp_run: message received, procId=%d, "
                "cmd=%d \n", procId, msg->cmd);

        switch (msg->cmd) {
            case SvrMsg_Cmd_START:
                msg->svrProcId = MultiProc_self();
                /* send message back */
                MessageQ_put(queId, (MessageQ_Msg)msg);
                break;

            case SvrMsg_Cmd_PROCESS:
                break;

            case SvrMsg_Cmd_STOP:
                count++;
                /*
                 * When stop commands have been recieved from all procs
                 * terminate server
                 */
                if (count == numProcs) {
                    /* Free last message */
                    MessageQ_free((MessageQ_Msg)msg);
                    running = FALSE;
                }
                break;

            default:
                System_printf("SvrDsp_run: unknown message, cmd=%d \n",
                        (msg->cmd));
                break;
        }


        /* decrement the done flag */
        if (Semaphore_getCount(SvrDsp_doneFlag) > 0) {
            Semaphore_pend(SvrDsp_doneFlag, BIOS_WAIT_FOREVER);
        }

    } /* while (running) */

leave:
    System_printf("<-- SvrDsp_run: %d \n", status);
    return(status);
}

/*
 *  ======== SvrDsp_setup ========
 */
Int SvrDsp_setup(Void)
{
    Int                 status = 0;
    Error_Block         eb;
    Int                 count;
    Semaphore_Params    semParams;
    Registry_Result     result;
    MessageQ_Params     msgqParams;
    UInt16              procId;
    Char                cbuf[48];

    /* TODO enter gate */

    /* initialize module state */
    Error_init(&eb);
    Module.localMsgQue = NULL;

    /* initialize the done flag */
    Semaphore_Params_init(&semParams);
    semParams.mode = Semaphore_Mode_COUNTING;
    count = MultiProc_getNumProcessors() * SvrMsg_NUM_MSGS;

    SvrDsp_doneFlag = Semaphore_create(count, &semParams, &eb);

    if (Error_check(&eb)) {
        status = -2;
        goto leave;
    }

    /* register with xdc.runtime to get a diags mask */
    result = Registry_addModule(&Registry_CURDESC, MODULE_NAME);

    if ((result != Registry_SUCCESS) && (result != Registry_ALREADY_ADDED)) {
        status = -1;
        goto leave;
    }

    /* enable some trace */
    Diags_setMask(MODULE_NAME"+FEX");
    System_printf("--> SvrDsp_setup:\n");

    /* compute the server queue name */
    procId = MultiProc_self();
    System_sprintf(cbuf, SvrMsg_SvrQueNameFmt, MultiProc_getName(procId));

    /* create local message queue (inbound messages) */
    MessageQ_Params_init(&msgqParams);
    Module.localMsgQue = MessageQ_create(cbuf, &msgqParams);

    if (Module.localMsgQue == NULL) {
        System_printf("SvrDsp_setup: message queue create failed\n");
        goto leave;
    }

    System_printf("SvrDsp_setup: slave is ready\n");

leave:

    /* TODO leave gate */
    System_printf("<-- SvrDsp_setup:\n");
    return(status);
}
